/****************************************************************************
debug_opr.c - Debugging functions for NRIA

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This file contains functions to dump the contents of selected NRIA
internal data structures in human-readable form.  Its primary purpose
is debugging NRIA and supporting routines.  For most of these
routines, pointers (except to character strings) are printed out but
not followed; that is the responsibility of whoever calls these
functions.  However, functions ending in "_tree" do follow pointers
and print out all the substructures pointed to.  Additional structures
will be selected for inclusion in this file whenever someone needs to
print them out and writes the necessary function to dump them.
****************************************************************************/

static const char rcsid[] = "$Id: debug_opr.c,v 1.6 1999/07/07 18:01:06 hughett Exp $ Release $Name: nria_1_0_6 $";


#include <stdio.h>
#include <stdlib.h>

#include "input.h"
#include "output.h"
#include "nria_lib.h"


void dump_string(FILE *out, char *str);
void dump_pointer(FILE *out, void *ptr);

void dump_output_file_rec(FILE *out, struct output_file_rec *rec);
void dump_input_file_rec(FILE *out, struct input_file_rec *rec);
void dump_output_hdr_rec(FILE *out, struct output_hdr_rec *rec);
void dump_hdr_rec(FILE *out, struct hdr_rec *rec);
void dump_slice_info_rec(FILE *out, struct slice_info_rec *rec);
void dump_slice_info_rec_tree(FILE *out, struct slice_info_rec *rec);
void dump_data_slice_rec(FILE *out, struct data_slice_rec *rec);
void dump_histo_params_rec(FILE *out, struct histo_params_rec *rec);
void dump_slice_roi_win_rec_tree(FILE *out, struct slice_roi_win_rec *rec);
void dump_slice_roi_win_rec(FILE *out, struct slice_roi_win_rec *rec);
void dump_slice_roi_rec(FILE *out, struct slice_roi_rec *rec);
void dump_roi_rec(FILE *out, struct roi_rec *rec);
void dump_curve_2d_rec(FILE *out, struct curve_2d_rec *rec);
void dump_curve_rec(FILE *out, struct curve_rec *rec);
void dump_float_bound_rec(FILE *out, struct float_bound_rec *rec);
void dump_matrix_2d_rec(FILE *out, struct matrix_2d_rec *rec);


void 
dump_output_file_rec(
   FILE *out,
   struct output_file_rec *rec)
{  fprintf(out, "\n*** Dumping an output_file_rec at ");
   dump_pointer(out, rec);
   if (rec == NULL) return;
   fprintf(out, "   file -> input_file_rec at ");
   dump_pointer(out, rec->file);
   fprintf(out, "   status = 0x%hx\n", rec->status);
   fprintf(out, "   dirfn = ");
   dump_string(out, rec->dirfn);
   fprintf(out, "   fn = ");
   dump_string(out, rec->fn);
   fprintf(out, "   dat_fn_suffix = ");
   dump_string(out, rec->dat_fn_suffix);
   fprintf(out, "   bin_fn_suffix = ");
   dump_string(out, rec->bin_fn_suffix);
   fprintf(out, "   hdr = output_hdr_rec at ");
   dump_pointer(out, &rec->hdr);
   dump_output_hdr_rec(out, &rec->hdr);
   return;
}


void
dump_input_file_rec(
   FILE *out,
   struct input_file_rec *rec)
{  fprintf(out, "\n*** Dumping an input_file_rec at ");
   dump_pointer(out, rec);
   if (rec == NULL) return;
   fprintf(out, "   prog -> prog_rec at ");
   dump_pointer(out, rec->prog);
   fprintf(out, "   status = 0x%hx\n", rec->status);
   fprintf(out, "   dirfn = ");
   dump_string(out, rec->dirfn);
   fprintf(out, "   fn = ");
   dump_string(out, rec->fn);
   fprintf(out, "   hdr_fn_suffix = ");
   dump_string(out, rec->hdr_fn_suffix);
   fprintf(out, "   img_fn_suffix = ");
   dump_string(out, rec->img_fn_suffix);
   fprintf(out, "   hdr = hdr_rec at ");
   dump_pointer(out, &rec->hdr);
   fprintf(out, "   img_fid -> FILE at ");
   dump_pointer(out, rec->img_fid);
   fprintf(out, "   img = img_rec at ");
   dump_pointer(out, &rec->img);
   fprintf(out, "   output = output_file_rec at ");
   dump_pointer(out, &rec->output);
   fprintf(out, "   slice_info = slice_info_rec at ");
   dump_pointer(out, &rec->slice_info);
   fprintf(out, "   histo_params = histo_params_rec at ");
   dump_pointer(out, &rec->histo_params);
   fprintf(out, "   pixel_type = pixel_type_rec at ");
   dump_pointer(out, &rec->pixel_type);
   fprintf(out, "   color_guide -> intensity_rec at ");
   dump_pointer(out, rec->color_guide);
   fprintf(out, "   edit_begin = ");
   dump_string(out, rec->edit_begin);
   fprintf(out, "   txt_format_fn = ");
   dump_string(out, rec->txt_format_fn);
   fprintf(out, "   count_format_fn = ");
   dump_string(out, rec->count_format_fn);
   fprintf(out, "   text = text_output_rec at ");
   dump_pointer(out, &rec->text);
   fprintf(out, "   count = count_analysis_rec at ");
   dump_pointer(out, &rec->count);
   fprintf(out, "   fco = file_co_register_rec at ");
   dump_pointer(out, &rec->fco);
   return;
}


void 
dump_output_hdr_rec(
   FILE *out,
   struct output_hdr_rec *rec)
{  
   int i;

   fprintf(out, "\n*** Dumping an output_hdr_rec at ");
   dump_pointer(out, rec);
   if (rec == NULL) return;
   fprintf(out, "   size = %ld\n", (long)rec->size);
   fprintf(out, "   id =");
   dump_string(out, rec->id);
   fprintf(out, "   rater = ");
   dump_string(out, rec->rater);
   fprintf(out, "   offset = output_field_offset_rec at ");
   dump_pointer(out, &rec->offset);
   fprintf(out, "   info = output_hdr_info_rec at ");
   dump_pointer(out, &rec->info);
   fprintf(out, "   actual_total_slice_rois = %lu\n", 
      rec->actual_total_slice_rois);
   fprintf(out, "   num_slice_rois = %lu\n", rec->num_slice_rois);
   fprintf(out, "   slice_nums -> long at ");
   dump_pointer(out, rec->slice_nums);
   for (i = 0; i < rec->num_slice_rois; i++ ) {
      fprintf(out, "   slice_nums[%d] = %ld\n", i, rec->slice_nums[i]);
   }

   return;
}

void
dump_hdr_rec(
   FILE *out,
   struct hdr_rec *rec)
{  fprintf(out, "\n*** Dumping a hdr_rec at ");
   dump_pointer(out, rec);
   if (rec == NULL) return;
   fprintf(out, "   size = %lu\n", rec->size);
   fprintf(out, "   id = ");
   dump_string(out, rec->id);
   fprintf(out, "   offset = input_field_offset_rec at ");
   dump_pointer(out, &rec->offset);
   fprintf(out, "   status = 0x%hx\n", rec->status);
   fprintf(out, "   data_info = data_info_rec at ");
   dump_pointer(out, &rec->data_info);
   fprintf(out, "   analysis = hdr_analysis_rec at ");
   dump_pointer(out, &rec->analysis);
   fprintf(out, "   num_scans = %lu\n", rec->num_scans);
   fprintf(out, "   scans -> hdr_scan_entry_rec at ");
   dump_pointer(out, rec->scans);
   fprintf(out, "   extra_info = extra_file_info_rec at ");
   dump_pointer(out, &rec->extra_info);
   fprintf(out, "   comment_size = %ld\n", rec->comment_size);
   fprintf(out, "   comment = ");
   dump_string(out, rec->comment);
   fprintf(out, "   user_defined_size = %ld\n", rec->user_defined_size);
   fprintf(out, "   user_defined = ");
   dump_string(out, (char *)rec->user_defined);
   return;
}


void
dump_slice_info_rec(
   FILE *out,
   struct slice_info_rec *rec)
{  fprintf(out, "\n*** Dumping a slice_info_rec at ");
   dump_pointer(out, rec);
   if (rec == NULL) return;
   fprintf(out, "   status = 0x%hx\n", rec->status);
   fprintf(out, "   max_sroi_wins = %lu\n", rec->max_sroi_wins);
   fprintf(out, "   num_sroi_wins = %lu\n", rec->num_sroi_wins);
   fprintf(out, "   sroi_wins -> array of slice_roi_win_rec at ");
   dump_pointer(out, rec->sroi_wins);
   fprintf(out, "   num_slices = %lu\n", rec->num_slices);
   fprintf(out, "   slices -> slice_attr_rec at ");
   dump_pointer(out, rec->slices);
   fprintf(out, "   min_pixel = %f\n",  rec->min_pixel);
   fprintf(out, "   max_pixel = %f\n",  rec->max_pixel);
   fprintf(out, "   total_3ds = %lu\n", rec->total_3ds);
   fprintf(out, "   num_3ds = %lu\n", rec->num_3ds);
   fprintf(out, "   _3ds -> array of buf_3d_rec at ");
   dump_pointer(out, rec->_3ds);
   fprintf(out, "   select_3d_gravity = %d\n", rec->select_3d_gravity);
   fprintf(out, "   translate -> vector_3d_rec at ");
   dump_pointer(out, &rec->translate);
   fprintf(out, "   rotate -> vector_3d_rec at ");
   dump_pointer(out, &rec->rotate);
   fprintf(out, "   group_num = %d\n", rec->group_num);
   fprintf(out, "   num_3d_sampled_pts = %ld\n", rec->num_3d_sampled_pts);
   fprintf(out, "   sampled_pts[...] omitted...\n");
   fprintf(out, "   sampled_pt0_base omitted...\n");
   fprintf(out, "   total_views = %lu\n", rec->total_views);
   fprintf(out, "   num_views = %lu\n", rec->num_views);
   fprintf(out, "   views -> array of *view_rec at ");
   dump_pointer(out, rec->views);
   fprintf(out, "   total_data_slices = %lu\n", rec->total_data_slices);
   fprintf(out, "   num_data_slices = %lu\n", rec->num_data_slices);
   fprintf(out, "   data_slices -> array of *data_slice_rec at ");
   dump_pointer(out, rec->data_slices);

   return;
}


void
dump_slice_info_rec_tree(
   FILE *out,
   struct slice_info_rec *rec)
{
   int i;

   dump_slice_info_rec(out, rec);

   /* FIXME: Contrary to what you might expect, the number of actual
      elements in sroi_wins appears to set by max_sroi_wins rather than
      num_sroi_wins; the role of num_sroi_wins is not known at this
      time. -PWH 19981123 */
   for (i = 0; i < rec->max_sroi_wins; i++)
       dump_slice_roi_win_rec_tree(out, rec->sroi_wins[i]);

   fprintf(out, "\n");
   fprintf(out, "*** skipping slice_attr_rec records...\n");
   fprintf(out, "*** skipping buf_3d_rec records...\n");
   fprintf(out, "*** skipping vector_3d_rec records...\n");
   fprintf(out, "*** skipping view_rec records...\n");
   
   for (i = 0; i < rec->num_data_slices; i++)
      dump_data_slice_rec(out, rec->data_slices[i]);

   return;
}


void
dump_data_slice_rec(
   FILE *out,
   struct data_slice_rec *rec)
{  fprintf(out, "\n*** Dumping a data_slice_rec at ");
   dump_pointer(out, rec);
   if (rec == NULL) return;

   fprintf(out, "   slice = hdr_scan_entry_rec at ");
   dump_pointer(out, &rec->slice);
   fprintf(out, "   attr = slice_attr_rec at ");
   dump_pointer(out, &rec->attr);
   fprintf(out, "   slice_roi_win = slice_roi_win_rec at ");
   dump_pointer(out, &rec->slice_roi_win);

   return;
}



void
dump_histo_params_rec(
   FILE *out,
   struct histo_params_rec *rec)
{  fprintf(out, "\n*** Dumping a histo_params_rec at ");
   dump_pointer(out, rec);
   if (rec == NULL) return;

   fprintf(out, "   file -> input_file_rec at ");
   dump_pointer(out, rec->file);
   fprintf(out, "   status2 = 0x%hx\n", rec->status2);
   fprintf(out, "   prog_status2 = 0x%hx\n", rec->prog_status2);
   fprintf(out, "   slice_roi_win_2d = slice_roi_win_rec at ");
   dump_pointer(out, &rec->slice_roi_win_2d);
   fprintf(out, "   slice_roi_win_1d = slice_roi_win_rec at ");
   dump_pointer(out, &rec->slice_roi_win_1d);
   fprintf(out, "   total_centers = %u\n", rec->total_centers);
   fprintf(out, "   cur_center = %u\n", rec->cur_center);
   fprintf(out, "   num_centers = %u\n", rec->num_centers);
   fprintf(out, "   centers -> pixel_center_rec at ");
   dump_pointer(out, rec->centers);
   fprintf(out, "   mid = %d\n", rec->mid);
   fprintf(out, "   range = %d\n", rec->range);
   fprintf(out, "   total_types = %u\n", rec->total_types);
   fprintf(out, "   cur_type = %u\n", rec->cur_type);
   fprintf(out, "   num_types = %u\n", rec->num_types);
   fprintf(out, "   types -> pixel_tissue_type_rec at ");
   dump_pointer(out, rec->types);
   fprintf(out, "   pixel_type_order = strs_rec at ");
   dump_pointer(out, &rec->pixel_type_order);
   fprintf(out, "   _2d_status = %hu\n", rec->_2d_status);
   fprintf(out, "   _2d = buf_2d_rec at ");
   dump_pointer(out, &rec->_2d);
   fprintf(out, "   map = buf_2d_rec at ");
   dump_pointer(out, &rec->map);
   fprintf(out, "   _1d_status = %hu\n", rec->_1d_status);
   fprintf(out, "   _2d_1d = buf_2d_rec at ");
   dump_pointer(out, &rec->_2d_1d);
   fprintf(out, "   cnt_1d = buf_2d_rec at ");
   dump_pointer(out, &rec->cnt_1d);
   fprintf(out, "   histo_1d = buf_2d_rec at ");
   dump_pointer(out, &rec->histo_1d);
   
   return;
}


/* Dump a slice_roi_win_rec and the substructures it points to */
void
dump_slice_roi_win_rec_tree(
   FILE *out,
   struct slice_roi_win_rec *rec)
{
   struct slice_roi_rec *slice_roi_1;
   struct roi_rec *roi_1;
   struct curve_2d_rec *c2dr;
   struct curve_rec *crv;
   struct float_xy_rec *xy;
   int iroi, ipen, icrv, ixy;

   dump_slice_roi_win_rec(out, rec);
   slice_roi_1 = &rec->slice_roi;
   dump_slice_roi_rec(out, slice_roi_1);
   for (iroi = 0; iroi < slice_roi_1->num_rois; iroi++) {
      roi_1 = slice_roi_1->rois[iroi];
      dump_roi_rec(out, roi_1); 
      for (ipen = 0; ipen < roi_1->num_pens; ipen++) {
	 c2dr = roi_1->pens[ipen];
         dump_curve_2d_rec(out, c2dr); 
         for (icrv = 0; icrv < c2dr->num_curves; icrv++) {
	    crv = c2dr->curves[icrv];
	    dump_curve_rec(out, crv);
	    dump_matrix_2d_rec(out, &crv->matrix);
	    dump_matrix_2d_rec(out, &crv->inv);
	    dump_float_bound_rec(out, &crv->fl_bound);
	    fprintf(out, "\n*** Dumping xy coordinate array at ");
	    xy = crv->xys;
	    dump_pointer(out, xy);
	    for (ixy = 0; ixy < crv->num_xys; ixy++)
	       fprintf(out, "   xy[%d] = %10.5f, %10.5f\n",
		       ixy, xy[ixy].x, xy[ixy].y);
	 }
      } 
   }
   return;
}


void
dump_slice_roi_win_rec(
   FILE *out,
   struct slice_roi_win_rec *rec)
{  int i;

   fprintf(out, "\n*** Dumping a slice_roi_win_rec at ");
   dump_pointer(out, rec);
   if (rec == NULL) return;

   fprintf(out, "   slice_roi = slice_roi_rec at  ");
   dump_pointer(out, &rec->slice_roi);
   fprintf(out, "   num_wins = %u\n", rec->num_wins);
   fprintf(out, "   wins -> image_display_window array at ");
   dump_pointer(out, (void *)rec->wins);
   for (i = 0; i < rec->num_wins; i++ ) {
      fprintf(out, "   wins[%d] -> image_display_window at ", i);
      dump_pointer(out, (void *)rec->wins[i]); }

   return;
}



void
dump_slice_roi_rec(
   FILE *out,
   struct slice_roi_rec *rec)
{  int i;

   fprintf(out, "\n*** Dumping a slice_roi_rec at ");
   dump_pointer(out, rec);
   if (rec == NULL) return;

   fprintf(out, "   status = 0x%hx\n", rec->status);
   fprintf(out, "   prog_status = 0x%hx\n", rec->prog_status);
   fprintf(out, "   xpixel_size = %f\n", rec->xpixel_size);
   fprintf(out, "   ypixel_size = %f\n", rec->ypixel_size);
   fprintf(out, "   dimx = %lu\n", rec->dimx);
   fprintf(out, "   dimy = %lu\n", rec->dimy);
   fprintf(out, "   total_rois = %lu\n", rec->total_rois);
   fprintf(out, "   cur_roi = %lu\n", rec->cur_roi);
   fprintf(out, "   num_rois = %lu\n", rec->num_rois);
   fprintf(out, "   rois -> pointer array at ");
   dump_pointer(out, (void *)rec->rois);
   for (i = 0; i < rec->num_rois; i++) {
      fprintf(out, "      rois [%d] -> roi_rec at ", i);
      dump_pointer(out, (void *)rec->rois[i]);  }

   return;
}



void
dump_roi_rec(
   FILE *out,
   struct roi_rec *rec)
{  int i;

   fprintf(out, "\n*** Dumping an roi_rec at ");
   dump_pointer(out, rec);
   if (rec == NULL) return;

   fprintf(out, "   label = ");
   dump_string(out, rec->label);
   fprintf(out, "   status = 0x%hx\n", rec->status);
   fprintf(out, "   prog_status = %hu\n", rec->prog_status);
   fprintf(out, "   total_pens = %lu\n", rec->total_pens);
   fprintf(out, "   cur_pen = %lu\n", rec->cur_pen);
   fprintf(out, "   num_pens = %lu\n", rec->num_pens);
   fprintf(out, "   pens -> pointer array at ");
   dump_pointer(out, (void *)rec->pens);
   for (i = 0; i < rec->num_pens; i++) {
      fprintf(out, "      pens[%d] -> curve_2d_rec at ", i);
      dump_pointer(out, rec->pens[i]); }
   fprintf(out, "   user = user_defined_rec at ");
   dump_pointer(out, &rec->user);

   return;
}



void
dump_curve_2d_rec(
   FILE *out,
   struct curve_2d_rec *rec)
{  int i;

   fprintf(out, "\n*** Dumping a curve_2d_rec at ");
   dump_pointer(out, rec);
   if (rec == NULL) return;

   fprintf(out, "   pen_type = %hu\n", rec->pen_type);
   fprintf(out, "   total_curves = %lu\n", rec->total_curves);
   fprintf(out, "   cur_curve = %lu\n", rec->cur_curve);
   fprintf(out, "   num_curves = %lu\n", rec->num_curves);
   fprintf(out, "   curves -> pointer array at ");
   dump_pointer(out, (void *)rec->curves);
   for ( i = 0; i < rec->num_curves; i++) {
      fprintf(out, "      curves[%d] -> curve_rec at ", i);
      dump_pointer(out, rec->curves[i]); }

   return;
}



void
dump_curve_rec(
   FILE *out,
   struct curve_rec *rec)
{  
   fprintf(out, "\n*** Dumping a curve_rec at ");
   dump_pointer(out, rec);
   if (rec == NULL) return;

   fprintf(out, "   status = 0x%hx\n", rec->status);
   fprintf(out, "   more_status = 0x%hx\n", rec->more_status);
   fprintf(out, "   total_xys = %lu\n", rec->total_xys);
   fprintf(out, "   num_xys = %lu\n", rec->num_xys);
   fprintf(out, "   fl_bound = float_bound_rec at ");
   dump_pointer(out, &rec->fl_bound);
   fprintf(out, "   xys -> float_xy_rec array at ");
   dump_pointer(out, rec->xys);
   fprintf(out, "   matrix = matrix_2d_rec at ");
   dump_pointer(out, &rec->matrix);
   fprintf(out, "   inv = matrix_2d_rec at ");
   dump_pointer(out, &rec->inv);

   return;
}



void
dump_float_bound_rec(
   FILE *out,
   struct float_bound_rec *rec)
{  
   fprintf(out, "\n*** Dumping a float_bound_rec at ");
   dump_pointer(out, rec);
   if (rec == NULL) return;

   fprintf(out, "   left   = %10.5f\n", rec->left);
   fprintf(out, "   top    = %10.5f\n", rec->top);
   fprintf(out, "   right  = %10.5f\n", rec->right);
   fprintf(out, "   bottom = %10.5f\n", rec->bottom);

   return;
}



void
dump_matrix_2d_rec(
   FILE *out,
   struct matrix_2d_rec *rec)
{  int i;

   fprintf(out, "\n*** Dumping a matrix_2d_rec at ");
   dump_pointer(out, rec);
   if (rec == NULL) return;

   fprintf(out, "   status = 0x%lx\n", rec->status);
   fprintf(out, "   buf = float array at ");
   dump_pointer(out, &rec->buf);
   for (i = 0; i < 3; i++)
      fprintf(out, "      %10.5f %10.5f %10.5f\n",
         rec->buf[i][0], rec->buf[i][1], rec->buf[i][2]); 

   return;
}

   


/* Dump a character string, which might be null or garbage */
void 
dump_string(
   FILE *out,
   char *str)
{  if (str == NULL) fprintf(out, "(null)\n");
   else             fprintf(out, "\"%.64s\"\n", str);
}


/* Dump a pointer, which might be null or garbage */
void 
dump_pointer(
   FILE *out,
   void *ptr)
{  if (ptr == NULL) fprintf(out, "(null)\n");
   else             fprintf(out, "%p\n", ptr);
}
