/****************************************************************************
bbl_dicom.c - ???

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.
****************************************************************************/

static const char rcsid[] = "$Id: bbl_dicom.c,v 1.9 1999/07/07 18:07:13 hughett Exp $ Release $Name: nria_1_0_6 $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>

#include "bbl_dicom.h"


#if 0
# define TRACE(x) x; fflush( stderr );
#else
# define TRACE(x)
#endif

#ifndef SEEK_SET
#define SEEK_SET 0
#define SEEK_CUR 1
#define SEEK_END 2
#endif


/* Functions defined in this file */
void initMessage(MessagePtr message);
MessagePtr msgOpen(char *fileName, int mode);
void msgClose(MessagePtr message);
const ElementValue getValue(MessagePtr message, ElementPtr element);
ElementPtr addElement(MessagePtr message, ElementGroupId elementGroupId,
    ElementId elementId, ElementLength elementLength,
    ElementValueOffset valueOffset);
static void addElementGroup(MessagePtr message, ElementGroupId elementGroupId,
    ElementPtr firstElement);
ElementPtr dupElement(ElementPtr element);
void freeElement(ElementPtr element);
ElementPtr findElement(MessagePtr message, ElementGroupId elementGroupId,
    ElementId elementId);
static ElementGroupPtr findElementGroup(MessagePtr message,
    ElementGroupId elementGroupId);
int getElements(MessagePtr message);
void dumpElements(MessagePtr message);
void dumpElement(MessagePtr message, ElementGroupId groupId, ElementId id);
const char *elementToString(MessagePtr message, ElementPtr element);



/*--------------------------------------------------------------------------
initMessage - ???
--------------------------------------------------------------------------*/

void
initMessage( MessagePtr message )
{
  message->error = 0;
  message->fileName[0] = '\0';
  message->fid = NULL;
  message->transferSyntax = 0;
  message->numElementGroups = 0;
  message->numElements = 0;
  message->firstGroup = NULL;
  message->lastGroup = NULL;
} /* initMessage() */



/*--------------------------------------------------------------------------
msgOpen - ???

msgOpen, version 0.1: assume mode = R
--------------------------------------------------------------------------*/

MessagePtr
msgOpen( char * fileName,
	 int mode )
{
  unsigned short	currGid, currEid;
  unsigned long		elemLen;
  struct stat		fileStat;
  MessagePtr		message = NULL;
  FILE			*fid;


  if ((mode & MSG_MODE) == MSG_READ)
  {
    if (stat( fileName, &fileStat ))
    {
      fprintf( stderr, "stat() error on \"%s\"\n", fileName );
      return NULL;
    }
  }
    
  if ((message = (MessagePtr) malloc( sizeof( Message ) )) == NULL)
    return NULL;

  initMessage( message );
  strcpy( message->fileName, fileName );

  if ((mode & MSG_MODE) == MSG_READ)
  {
    if ((fid = fopen( fileName, "r" )) == NULL)
    {
      fprintf( stderr, "fopen() error on \"%s\"\n", fileName );
      free( message );
      return NULL;
    }
    message->size = fileStat.st_size;
  }
    
  fseek( fid, 0, SEEK_SET );
  message->status |= mode | MSG_OPEN;		/* open */
  message->fid = fid;
    
    
  if ((mode & MSG_MODE) == MSG_READ)
  {
    /*
     *  If we read in an Element (ggggH.0000H), the Length should be
     *    4, and its Value should be a 32-bit double word, which is the
     *    even number of bytes from the end of this Element's Value
     *    field to the beginning of the next Group.
     *
     *  If we read in an Element (ggggH.0000H), and the Length is not
     *    4, we assume the data is in okay condition, but is stored
     *    in Little-Endian format instead of Big-Endian format.
     *
     *  If we don't have access to a Group Length Element, then we
     *    have to guess.  In this case, we choose the default of
     *    Little-Endian.  (Change from previous versions of DICOM
     *    standard: the Group Length Element is now optional)
     */
    if ((fread( &currGid, sizeof( currGid ), 1, fid ) != 1) ||
	(fread( &currEid, sizeof( currEid ), 1, fid ) != 1) ||
	(fread( &elemLen, sizeof( elemLen ), 1, fid ) != 1))
    {
      fprintf( stderr, "Read error: (%04x.%04x), el.len = %ld\n",
	      currGid, currEid, elemLen );
      msgClose( message );
      exit( 1 );
    }

    message->transferSyntax = LITTLE_ENDIAN;
    if (currEid == 0)
    {
      if (elemLen == 0x00000004)
	message->transferSyntax = BIG_ENDIAN;
      else if (elemLen != 0x04000000)
      {
	fprintf( stderr, "Warning: garbled Group Length Element (%08lx).\n",
		 elemLen );
	fprintf( stderr, "  (Should be either 00000004 or 04000000)\n" );
	fprintf( stderr, "  Assuming transfer syntax is Little Endian.\n\n" );
      }
    }
    else
    {
      fprintf( stderr, "Warning: can't find a Group Length Element.\n" );
      fprintf( stderr, "  Assuming transfer syntax is Little Endian.\n\n" );
    }
    TRACE( fprintf( stderr, "Transfer syntax: %s\n",
		   (message->transferSyntax == BIG_ENDIAN ?
		    "Big Endian" : "Little Endian" ) ); );
  }

  return message;

} /* msgOpen() */



/*--------------------------------------------------------------------------
msgClose - ???
--------------------------------------------------------------------------*/

void
msgClose( MessagePtr message )
{
  if (message)
  {
    fclose( message->fid );
    message->status &= ~MSG_OPENED;
    message->error = 0;
    message->fid = NULL;
  }
} /* msgClose() */



/*--------------------------------------------------------------------------
getValue - ???
--------------------------------------------------------------------------*/

CONST ElementValue
getValue( MessagePtr message, ElementPtr element )
{
  static ElementValue	currValue;
  static size_t		currSize;
  DefinitionPtr		currDefn;
  FILE			*fid;
  off_t			currOffset;
  unsigned short	*uShortPtr;
  int			counter;


  currDefn = findElementDefn( element->groupId, element->id );

  fid = message->fid;
  currOffset = ftell( fid );
  if (currValue)
  {
    if (currSize < element->length)
    {
      if ((currValue = (ElementValue) realloc( currValue, element->length ))
	  == NULL)
      {
	free( currValue );
	currValue = NULL;
	currSize = 0;
	return NULL;
      }
      currSize = element->length;
    }
    else
    {
      memset( currValue, 0, currSize );
    }
  }
  else if ((currValue = (ElementValue) malloc( element->length )) == NULL)
  {
    return NULL;
  }
  else
  {
    currSize = element->length;
  }

  switch (currDefn->valueRepresentation)
  {

  case ELEM_VR_OW:
    element->valueSize = element->length / sizeof( unsigned short );
    fseek( fid, element->value.offset, SEEK_SET );
    fread( currValue, element->length, 1, fid );
    fseek( fid, currOffset, SEEK_SET );
    if (message->transferSyntax == LITTLE_ENDIAN)
    {
      for (uShortPtr = (unsigned short *) currValue, counter = 0;
	   counter < element->valueSize;
	   uShortPtr ++, counter ++)
      {
	*(uShortPtr) = swapUShort( *(uShortPtr) );
      }
    }
    return currValue;

  case ELEM_VR_OB:
  case ELEM_VR_LT:
  case ELEM_VR_ST:
    element->valueSize = element->length;
    fseek( fid, element->value.offset, SEEK_SET );
    fread( currValue, element->length, 1, fid );
    fseek( fid, currOffset, SEEK_SET );
    return currValue;
  }

  return currValue;
} /* getValue */



/*--------------------------------------------------------------------------
addElement - ???

Add element to internal representation of message.  If we don't
understand this element (can't find the definition), do not add it to
message.
--------------------------------------------------------------------------*/

ElementPtr
addElement( MessagePtr message,
	    ElementGroupId elementGroupId,
	    ElementId elementId,
	    ElementLength elementLength,
	    ElementValueOffset valueOffset )
{
  ElementGroupPtr	elementGroup;
  ElementPtr		element;
  DefinitionPtr		elementDefn;
  off_t			currOffset;
  int			currLength;
  FILE			*fid;
  unsigned long		*uLongPtr;
  unsigned short	*uShortPtr;
  unsigned char		*uCharPtr;


  TRACE( fprintf( stderr, "addElement: message = %p\n", (void *) message );
	fprintf( stderr, "	(%04X.%04X): %d bytes, offset = %d\n",
		elementGroupId,
		elementId,
		elementLength,
		valueOffset ); );

  if (!message)
    return NULL;

  if ((elementDefn = findElementDefn( elementGroupId, elementId )) == NULL)
    return NULL;

  if ((element = (ElementPtr) malloc( sizeof( Element ) )) == NULL)
    return NULL;


  fid = message->fid;
  currOffset = ftell( fid );

  switch (VALREP( elementDefn ))
  {

    /*
     *  Byte order insensitive binary types
     */

  case ELEM_VR_UL:		/* Unsigned Long (32-bit binary integers) */

    if (VALMULT( elementDefn ) == ELEM_VM_1_THRU_N)
    {
      element->multiplicity = elementLength / sizeof( unsigned long );
    }
    else
      element->multiplicity = VALMULT( elementDefn );
    if ((uLongPtr = (unsigned long *) malloc( elementLength )) == NULL)
      return NULL;
    fseek( fid, valueOffset, SEEK_SET );
    fread( uLongPtr, sizeof( unsigned long ), element->multiplicity, fid );
    fseek( fid, currOffset, SEEK_SET );

    if (message->transferSyntax == LITTLE_ENDIAN)
    {
      for (currLength = 0; currLength < element->multiplicity; currLength ++)
      {
	*(uLongPtr + currLength) = swapULong( *(uLongPtr + currLength) );
      }
    }

    element->valueLocation = ELEM_VL_MEM;
    element->valueSize = elementLength;
    element->value.ptr = (ElementValue) uLongPtr;
    break;

  case ELEM_VR_US:		/* Unsigned Short (16-bit binary integers) */

    if (VALMULT( elementDefn ) == ELEM_VM_1_THRU_N)
    {
      element->multiplicity = elementLength / sizeof( unsigned short );
    }
    else
      element->multiplicity = VALMULT( elementDefn );
    if ((uShortPtr = (unsigned short *) malloc( elementLength )) == NULL)
      return NULL;
    fseek( fid, valueOffset, SEEK_SET );
    fread( uShortPtr, sizeof( unsigned short ), element->multiplicity, fid );
    fseek( fid, currOffset, SEEK_SET );

    if (message->transferSyntax == LITTLE_ENDIAN)
      for (currLength = 0; currLength < element->multiplicity; currLength ++)
	*(uShortPtr + currLength) = swapUShort( *(uShortPtr + currLength) );

    element->valueLocation = ELEM_VL_MEM;
    element->valueSize = elementLength;
    element->value.ptr = (ElementValue) uShortPtr;
    break;


    /*
     *  Values which may be too large to store in memory.
     */

  case ELEM_VR_OB:		/* Other Byte String (8-bit integers,
				   Little/Big-Endian sensitive) */
  case ELEM_VR_OW:		/* Other Word String (16-bit integers,
				   Little/Big-Endian sensitive) */
  case ELEM_VR_LT:		/* Long Text */
  case ELEM_VR_ST:		/* Short Text */

    /*
     *  Data elements with a VR of SQ, OW, or OB shall always
     *    have a Value Multiplicity of one.  (NEMA PS3.5, Part 5, pg. 21)
     *
     *  We shall discover true multiplicity of Text formats when we
     *    need to access this data, otherwise save time by not bothering.
     */
    element->multiplicity = VALMULT( elementDefn );
    element->valueLocation = ELEM_VL_FILE;
    element->valueSize = elementLength;
    element->value.offset = valueOffset;
    break;


    /*
     *  String types
     */

  default:

    currLength = (elementLength <= MAX_STRING_ELEM_VALUE_LEN ?
		  elementLength : MAX_STRING_ELEM_VALUE_LEN);
    if ((uCharPtr = (unsigned char *) malloc( currLength + 1 )) == NULL)
      return NULL;
    fseek( fid, valueOffset, SEEK_SET );
    fread( uCharPtr, currLength, 1, fid );
    fseek( fid, currOffset, SEEK_SET );
    *(uCharPtr + currLength) = 0;
    element->valueLocation = ELEM_VL_MEM;
    element->valueSize = currLength;
    element->value.ptr = (ElementValue) uCharPtr;
    if (VALMULT( elementDefn ) == ELEM_VM_1_THRU_N)
    {
      element->multiplicity = 1;
      uCharPtr = (unsigned char *) element->value.ptr;
      while (uCharPtr = (unsigned char *)
	     strchr( (CONST char *) uCharPtr, '\\' ))
      {
	uCharPtr ++;
	element->multiplicity ++;
      }
    }
    else
      element->multiplicity = VALMULT( elementDefn );
    break;

  }

  element->groupId = elementGroupId;
  element->id = elementId;
  element->length = elementLength;
  element->nextElement = NULL;

  if ((elementGroup = findElementGroup( message, elementGroupId )) == NULL)
    addElementGroup( message, elementGroupId, element );
  else
  {
    elementGroup->lastElement->nextElement = element;
    elementGroup->lastElement = element;
  }

  message->numElements ++;

  return element;
} /* addElement() */



/*--------------------------------------------------------------------------
addElementGroup - ???
--------------------------------------------------------------------------*/

static void
addElementGroup( MessagePtr message,
		 ElementGroupId elementGroupId,
		 ElementPtr firstElement )
{
  ElementGroupPtr	newElementGroup;

  TRACE( fprintf( stderr, "addElementGroup: message = %p", (void *) message );
	fprintf( stderr, ", elementGroupId = %04X", elementGroupId );
	fprintf( stderr, ", firstElement = %p\n", (void *) firstElement ); );

  newElementGroup = (ElementGroupPtr) malloc( sizeof( ElementGroup ) );
  newElementGroup->groupId = elementGroupId;
  newElementGroup->firstElement = firstElement;
  newElementGroup->lastElement = firstElement;

  newElementGroup->nextGroup = NULL;

  if (message->firstGroup == NULL)
  {
    message->firstGroup = message->lastGroup = newElementGroup;
  }
  else
  {
    message->lastGroup->nextGroup = newElementGroup;
    message->lastGroup = newElementGroup;
  }

  message->numElementGroups ++;
} /* addElementGroup() */



/*--------------------------------------------------------------------------
dupElement - ???
--------------------------------------------------------------------------*/

ElementPtr
dupElement( ElementPtr element )
{
  ElementPtr		elementCopy;

  if ((elementCopy = (ElementPtr) malloc( sizeof( Element ) )) == NULL)
    return NULL;

  elementCopy->groupId = element->groupId;
  elementCopy->id = element->id;
  elementCopy->length = element->length;
  elementCopy->multiplicity = element->multiplicity;
  elementCopy->valueLocation = element->valueLocation;
  elementCopy->valueSize = element->valueSize;

  switch (element->valueLocation)
  {

  case ELEM_VL_MEM:
    if ((elementCopy->value.ptr = (ElementValue)
	 malloc( element->valueSize )) == NULL)
      return NULL;
    memcpy( elementCopy->value.ptr, element->value.ptr, element->valueSize );
    break;

  case ELEM_VL_FILE:
    elementCopy->value.offset = element->value.offset;
    break;

  }

  elementCopy->nextElement = NULL;

  return element;
} /* dupElement() */



/*--------------------------------------------------------------------------
freeElement - ???
--------------------------------------------------------------------------*/

void
freeElement( ElementPtr element )
{
  if (element)
  {
    if (element->valueLocation == ELEM_VL_MEM)
      free( element->value.ptr );
    free( element );
  }

  return;
} /* freeElement() */



/*--------------------------------------------------------------------------
findElement - ???
--------------------------------------------------------------------------*/

ElementPtr
findElement( MessagePtr message,
	     ElementGroupId elementGroupId,
	     ElementId elementId )
{
  ElementGroupPtr	elementGroup;
  ElementPtr	currElement;

  TRACE( fprintf( stderr, "findElement: message = %p", (void *) message );
	fprintf( stderr, ", elementGroupId = %04X, elementId = %04X\n",
		elementGroupId,
		elementId ); );

  if (!message)
  {
    fprintf( stderr, "findElement() called with NULL message.\n" );
    return NULL;
  }
  message->error = 0;
  if ((elementGroup = findElementGroup( message, elementGroupId )) == NULL)
    return NULL;		/*  error already set  */
  currElement = elementGroup->firstElement;
  /*
   *  The following comparison depends on the fact that the
   *    definitions are sorted in ascending order, by group id,
   *    then by element id, to improve search time.
   */
  while (currElement && (currElement->id < elementId))
    currElement = currElement->nextElement;

  if (!currElement || currElement->id > elementId)
  {
    message->error = ELEMENT_NOTFOUND;
    return NULL;
  }

  /*
   *  If we didn't find the definition, this returns NULL, otherwise we
   *    return a pointer to the element definition.
   */
  return currElement;
} /* freeElement() */



/*--------------------------------------------------------------------------
findElementGroup - ???
--------------------------------------------------------------------------*/

static ElementGroupPtr
findElementGroup( MessagePtr message,
		  ElementGroupId elementGroupId )
{
  ElementGroupPtr currElementGroup;

  TRACE( fprintf( stderr, "findElementGroup: message = %p", (void *) message );
	fprintf( stderr, ", elementGroupId = %04X\n", elementGroupId ); );

  if (!message)
    return NULL;

  currElementGroup = message->firstGroup;
  /*
   *  The following comparison depends on the fact that the
   *    definitions are sorted in ascending order, by group id,
   *    then by element id, to improve search time.
   */
  while (currElementGroup && (currElementGroup->groupId < elementGroupId))
    currElementGroup = currElementGroup->nextGroup;

  if (!currElementGroup || currElementGroup->groupId > elementGroupId)
  {
    message->error = ELEMENT_NOTFOUND;
    return NULL;
  }

  /*
   *  If we didn't find the definition, this returns NULL, otherwise we
   *    return a pointer to the group definition.
   */
  return currElementGroup;
} /* findElementGroup() */



/*--------------------------------------------------------------------------
getElements - ???
--------------------------------------------------------------------------*/

int
getElements( MessagePtr message )
{
  ushort		currGid, currEid;
  unsigned long		elemLen, offset = 0;
  FILE			*fid;
  ElementPtr		currElement = NULL;


  if (!message)
  {
    fprintf( stderr, "getElements() called with NULL message ptr.\n" );
    return INVALID_MESSAGE;
  }
  fid = message->fid;
  fseek( fid, 0, SEEK_SET );

  /*  Just so that we don't run off the end of the file... */
  while (offset < message->size)
  {

    /*
     *  DICOM Element structure:
     *
     * Offset	Size	Description
     * (bytes)	(bytes)
     *
     *	0	2 	Group ID (16-bit word, unsigned integer)
     *	2	2	Element ID (16-bit word, unsigned integer)
     *	4	4	Element Length (32-bit double word, unsigned integer)
     *	8	???	Element Value
     *
     *  Note that the size of the Element Value is determined by
     *    interpreting the Element Length as an unsigned integer
     *    (NEMA Standard 10-16-1985, Section 4.5, Paragraph 1)
     *  Note that all words are in Little-Endian format, the
     *    default Transfer Syntax for DICOM v3.0.
     */
    
    
    /*
     *  Read in the next Group ID, Element ID, and Element Length.
     */
    if ((fread( &currGid, sizeof( currGid ), 1, fid ) != 1) ||
	(fread( &currEid, sizeof( currEid ), 1, fid ) != 1) ||
	(fread( &elemLen, sizeof( elemLen ), 1, fid ) != 1))
    {
      fprintf( stderr, "Read error: (%04X.%04X), el.len = %ld\n",
	      currGid, currEid, elemLen );
      msgClose( message );
      return INVALID_MESSAGE;
    }
    if (message->transferSyntax == LITTLE_ENDIAN)
    {
      currGid = swapUShort( currGid );
      currEid = swapUShort( currEid );
      elemLen = swapULong( elemLen );
    }

    /*  The value is located 8 bytes into the element, since
	the header (gid, eid, and elen) take up 8 bytes.  */
    currElement = addElement( message, currGid, currEid, elemLen, offset+8 );

    /*
     *  Move on to the next element.  The magical number 8 is derived
     *    from the fact that the header of an Element is 8 bytes.
     */
    offset += elemLen + 8;
    fseek( fid, elemLen, SEEK_CUR );

  };  /*  While there is still more of the message  */

  return 0;

}  /*  getElements */



/*--------------------------------------------------------------------------
dumpElements - ???
--------------------------------------------------------------------------*/

void
dumpElements( MessagePtr message )
{
  ElementGroupPtr	currGroup;
  ElementPtr		currElement;
  DefinitionPtr		elementDefn;

  if (!message)
  {
    printf( "Empty message.\n" );
    return;
  }

  if (message->transferSyntax == BIG_ENDIAN)
    printf( "Transfer syntax: Big Endian\n" );
  else
    printf( "Transfer syntax: Little Endian\n" );

  currGroup = message->firstGroup;

  while (currGroup)
  {
    printf( "\n%04X", currGroup->groupId );
    currElement = currGroup->firstElement;

    while (currElement)
    {

      if (currElement->id == 0)
	printf( "      Length: %s bytes",
	       elementToString( message, currElement ) );
      else
      {
	elementDefn = findElementDefn( currElement->groupId,
				      currElement->id );
	printf( "\n%04X %04X ",
	       currElement->groupId,
	       currElement->id );
	printf( "%-32s ", elementDefn->elementName );
	printf( "%c%c % 3d  ",
	       elementDefn->valueRepresentation / 256,
	       elementDefn->valueRepresentation
	       - (elementDefn->valueRepresentation / 256) * 256,
	       currElement->multiplicity );
	printf( "%s", elementToString( message, currElement ) );
      }

      currElement = currElement->nextElement;
    }

    currGroup = currGroup->nextGroup;
  }

  printf( "\n***  End of message  ***\n" );

} /* dumpElements() */



/*--------------------------------------------------------------------------
dumpElement - ???
--------------------------------------------------------------------------*/

void
dumpElement( MessagePtr message, ElementGroupId groupId, ElementId id )
{
  ElementPtr		currElement;

  if (!message)
  {
    printf( "Empty message.\n" );
    return;
  }

  currElement = findElement( message, groupId, id );
  if ( currElement != 0 )
  {
    printf( "%s\n", elementToString( message, currElement ) );
  }
  else
  {
    printf( "NOT FOUND\n" );
  }

} /* dumpElement() */



/*--------------------------------------------------------------------------
elementToString - ???
--------------------------------------------------------------------------*/

CONST char *
elementToString( MessagePtr message, ElementPtr element )
{
  static char		currString[MAX_TEXT_ELEM_VALUE_LEN + 1];
  char			tempString[MAX_TEXT_ELEM_VALUE_LEN + 1];
  static int		currLength;
  long			currOffset;
  FILE			*fid;
  DefinitionPtr		elementDefn;


  if (!message)
  {
    TRACE( fprintf( stderr, "elementToString: empty message\n" ); );
    return NULL;
  }

  if (!element)
    return NULL;

  if ((elementDefn = findElementDefn( element->groupId, element->id )) == NULL)
    return NULL;

  switch (VALREP( elementDefn ))
  {


    /*
     *  Text types
     */

  case ELEM_VR_LT:		/* Long Text, 10240 chars max. */

  case ELEM_VR_ST:		/* Short Text, 1024 chars max. */

    TRACE( fprintf( stderr, "  --  --  --  Text type\n" ); );

    /* Currently, this only supports printing out the FIRST
     *  MAX_TEXT_ELEM_VALUE_LEN characters of the LT or ST.  */
    fid = message->fid;
    currOffset = ftell( fid );
    fseek( fid, element->value.offset, SEEK_SET );
    currLength = (element->length <= MAX_TEXT_ELEM_VALUE_LEN ?
		  element->length : MAX_TEXT_ELEM_VALUE_LEN);
    fread( currString, currLength, 1, fid );
    fseek( fid, currOffset, SEEK_SET );
    *(currString + currLength) = 0;
    break;


    /*
     *  Byte-order-sensitive types
     */
  case ELEM_VR_OB:		/* Other Byte String (8-bit integers,
				   Little/Big-Endian sensitive) */
    sprintf( currString, "(OBString, %ld bytes)", element->length );
    break;

  case ELEM_VR_OW:		/* Other Word String (16-bit integers,
				   Little/Big-Endian sensitive) */
    sprintf( currString, "(OWString, %ld bytes)", element->length );
    break;


    /*
     *  Binary integer types
     */

  case ELEM_VR_UL:	/* Unsigned Long (32-bit binary integers) */

    TRACE( fprintf( stderr, "  --  --  --  Unsigned long type\n" ); );

    *currString = '\0';
    for (currLength = 0; currLength < element->multiplicity; currLength ++)
    {
      sprintf( tempString, "%lu\\",
	      *((unsigned long *) element->value.ptr + currLength) );
      strcat( currString, tempString );
    }
    /*  Replace last '\' with NULL to end string.  */
    *(currString + (strlen( currString ) - 1)) = '\0';
    break;

  case ELEM_VR_US:	/* Unsigned Short (16-bit binary integers) */

    TRACE( fprintf( stderr, "  --  --  --  Unsigned short type\n" ); );

    *currString = '\0';
    for (currLength = 0; currLength < element->multiplicity; currLength ++)
    {
      sprintf( tempString, "%hu\\",
	      *((unsigned short *) element->value.ptr + currLength) );
      strcat( currString, tempString );
    }
    /*  Replace last '\' with NULL to end string.  */
    *(currString + (strlen( currString ) - 1)) = '\0';
    break;


    /*
     *  String types
     */

  default:

    return (char *) element->value.ptr;

  }

  return currString;
} /* elementToString() */


