/****************************************************************************
ndat2sroi.c - Extract slice ROI coordinates from an ndat file

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

Author:  Paul Hughett


This program extracts the slice ROI coordinates from an NRIA ndat file
and outputs them in a format easily readable by another computer
program or even a human.  See the man page for usage instructions.

****************************************************************************/

static const char rcsid[] = "$Id: ndat2sroi.c,v 1.6 1999/07/07 18:07:14 hughett Exp $ Release $Name: nria_1_0_6 $";

#include <stdio.h>
#include <stdlib.h>

#include "input.h"
#include "output.h"


/* FIXME:  This program needs to know all sorts of nria internals, which
   is ugly if feasible.  -PWH 19990706 */
#include "mri_format.h"
#include "nria_procedures.h"

/* FIXME:  These prototypes really ought to live in a header file somewhere,
   along with all the other NRIA library routines used in this file */
/***
extern void init_output_file_rec(
     struct output_file_rec *output,
     struct input_file_rec *file);
extern int rd_dat_file(char *fn, struct output_file_rec *output);
***/

/* Required by NRIA library if DEBUG defined; controls debugging output */
int debugFlag = 0;


/* ndat2sroi.c */
int main(int argc, char *argv[]);
static void dump_sroi_coords(struct output_file_rec *output);
void free_img_rec(struct img_rec *img);
void init_img_rec(struct img_rec *img, struct input_file_rec *file);



/*-------------------------------------------------------------------------
main - ???
-------------------------------------------------------------------------*/

int
main (int argc, char *argv[])
{  struct output_file_rec output;
   struct input_file_rec input;
   int stat;

   /* Initialize the data structures */
   init_output_file_rec(&output, &input);
   init_input_file_rec(&input, NULL);    /* FIXME:  Is NULL valid here? */

   /* Parse arguments to get the input file name */
   if (argc < 2) {
      fprintf(stderr, "Not enough arguments.\n");
      exit(EXIT_FAILURE); }

   /* Read the ndat file into core */
   stat = rd_dat_file(argv[1], &output);
   if (stat != 0) {
      fprintf(stderr, "Error reading the ndat file\n");
      exit(EXIT_FAILURE); }

   /* Look at selected contents from the ndat file */
   dump_sroi_coords(&output);

   return 0;
}



/*-------------------------------------------------------------------------
dump_sroi_coords - dump the coordinates of the slice roi's found.
-------------------------------------------------------------------------*/

static void
dump_sroi_coords(struct output_file_rec *output)
{  FILE *out = stdout;
   struct input_file_rec *pifr;
   struct slice_info_rec *psir;
   struct slice_roi_win_rec *srwr;
   struct slice_roi_rec *srr;
   struct roi_rec *roi;
   struct curve_2d_rec *c2dr;
   struct curve_rec *crv;
   struct float_xy_rec xy;
   int islice, iroi, ipen, icrv, ixy;
   char *label;

   /* Dig down to the slice_info_rec */
   pifr = output->file;          /* input_file_rec */
   psir = &pifr->slice_info;     /* slice_info_rec */

   /* Loop over the slice_roi_win_recs found.  For some unknown reason,
      the number of these is kept in max_sroi_wins, not num_sroi_wins. */
   for (islice = 0; islice < psir->max_sroi_wins; islice++) {
       srwr = psir->sroi_wins[islice];   /* slice_roi_win_rec */
       srr  = &srwr->slice_roi;          /* slice_roi_rec */
       

       /* Loop over the roi_recs in the slice_roi_rec */
       for (iroi = 0; iroi < srr->num_rois; iroi++) {
	   roi = srr->rois[iroi];         /* roi_rec */
	   label = roi->label;

	   /* Loop over the pens in the roi_rec */
	   for (ipen = 0; ipen < roi->num_pens; ipen++) {
	       c2dr = roi->pens[ipen];     /* curve_2d_rec */

	       /* Loop over the curves in the curve_2d_rec */
	       for (icrv = 0; icrv < c2dr->num_curves; icrv++) {
		   crv = c2dr->curves[icrv];   /* curve_rec */

		   /* Print the label and slice index */
		   fprintf(out, "\"%s\"  %d  {\n", label, islice);

		   /* Print coords for each point in the curve rec */
		   for (ixy = 0; ixy < crv->num_xys; ixy++) {
		       xy = crv->xys[ixy];   /* float_xy_rec */
		       fprintf(out, "   %f  %f\n", xy.x, xy.y);

		   }

		   /* Finish up printing this roi */
		   fprintf(out, "}\n");

	       }
	   }
       }
   }

   return;
}


/*--------------------------------------------------------------------------
free_img_rec - free an image record
init_img_rec - initialize an image record

These two routines are copied from the NRIA file img_buf_opr.c with
sufficient modifications so that they do not require XView or the GUI
to serve the present purpose.  In particular, free_img_rec is not
needed here, so it is just converted to a stub that will complain if
called; it is called from the NRIA source file input_file_rec.c, which
we need for other functions.  On the other hand, init_img_rec needs no
changes at all.
---------------------------------------------------------------------------*/

void
free_img_rec(
   struct img_rec *img)
{  
   fprintf(stderr, "The function free_img_rec is not yet implemented.\n");
   exit(EXIT_FAILURE);
}


void
init_img_rec(
   struct img_rec *img,
   struct input_file_rec *file)
{
   img->file = file;
   img->data = NULL;
   img->num_data = 0;
}
